const std = @import("std");

const BTree = struct {
    const Self = @This();

    const B: usize = 3;
    const CAPACITY: usize = 2 * B - 1;
    const NUM_EDGES: usize = 2 * B;

    ally: std.mem.Allocator = std.heap.c_allocator,
    root: ?NodeOrLeaf,

    fn create(ally: std.mem.Allocator) Self {
        return Self{
            .ally = ally,
            .root = null,
        };
    }

    fn insert(self: *Self, value: u32) !void {
        std.debug.print("attempting to insert {} into ", .{value});
        self.dbg();

        if (self.root) |*root| {
            const result = try root.as_leaf().insert_value(value);
            if (result) |split| {
                // create new node which will replace self.
                const parent = try Node.create(self.ally);
                parent.leaf.level = split.left.level + 1;
                parent.leaf.push_value(split.middle);
                parent.insert_node(NodeOrLeaf.from_leaf(split.left));
                parent.insert_node(NodeOrLeaf.from_leaf(split.right));

                self.root = .{ .internal = parent };
            }
        } else {
            var leaf: *Leaf = try self.ally.create(Leaf);
            errdefer self.ally.destroy(leaf);
            leaf.init(self.ally);
            leaf.push_value(value);
            self.root = NodeOrLeaf{ .leaf = leaf };
        }
    }

    fn dbg(self: *Self) void {
        if (self.root) |root| {
            root.dbg();
            std.debug.print("\n", .{});
        }
    }

    fn destroy(self: *Self) void {
        if (self.root) |*root| {
            root.destroy();
        }
    }

    const NodeOrLeafTag = enum {
        internal,
        leaf,
    };

    const NodeOrLeaf = union(NodeOrLeafTag) {
        internal: *Node,
        leaf: *Leaf,

        fn destroy(self: NodeOrLeaf) void {
            switch (self) {
                .internal => |node| {
                    node.destroy();
                },
                .leaf => |leaf| {
                    leaf.destroy();
                },
            }
        }

        fn as_leaf(self: NodeOrLeaf) *Leaf {
            switch (self) {
                .internal => |node| {
                    return node.as_leaf();
                },
                .leaf => |leaf| {
                    return leaf;
                },
            }
        }

        fn from_leaf(leaf: *Leaf) NodeOrLeaf {
            if (leaf.level == 0) {
                return .{ .leaf = leaf };
            } else {
                return .{ .internal = @ptrCast(*Node, leaf) };
            }
        }

        fn dbg(self: NodeOrLeaf) void {
            switch (self) {
                .internal => |node| {
                    node.dbg();
                },
                .leaf => |node| {
                    node.dbg();
                },
            }
        }
    };

    const Node = struct {
        leaf: Leaf,

        edges: [NUM_EDGES]?NodeOrLeaf = [_]?NodeOrLeaf{null} ** NUM_EDGES,

        fn create(ally: std.mem.Allocator) !*Node {
            var node = try ally.create(Node);
            node.init(ally);

            return node;
        }

        fn destroy(self: *Node) void {
            for (self.get_edges()) |edge| {
                if (edge) |*edg| {
                    edg.destroy();
                }
            }
            self.leaf.ally.destroy(self);
        }

        fn init(self: *Node, ally: std.mem.Allocator) void {
            self.* = Node{ .leaf = Leaf{ .ally = ally } };
        }

        fn as_leaf(self: *Node) *Leaf {
            return &self.leaf;
        }

        fn insert_node(self: *Node, child: NodeOrLeaf) void {
            const self_leaf = self.as_leaf();
            const ls = child.as_leaf().get_values()[0];

            var idx: u16 = self_leaf.len;
            for (self_leaf.get_values(), 0..) |v, i| {
                if (v > ls) {
                    idx = @intCast(u16, i);
                    break;
                }
            }

            if (self.get_edges()[idx]) |_| {
                std.debug.print("edge already present?:", .{});
                child.dbg();
                std.debug.print("\n", .{});
            } else {
                child.as_leaf().parent = .{ .parent = self, .idx = idx };
                self.get_edges()[idx] = child;
            }
        }

        fn get_edges(self: *Node) []?NodeOrLeaf {
            const len = self.leaf.len + 1;
            return self.edges[0..len];
        }

        fn dbg(self: *Node) void {
            const values = self.leaf.get_values();
            const edges = self.get_edges()[0..values.len];
            std.debug.print("{{ ", .{});
            for (values, edges) |v, e| {
                if (e) |edge| {
                    edge.dbg();
                    std.debug.print(", ", .{});
                }
                std.debug.print("{}, ", .{v});
            }
            if (self.get_edges()[values.len]) |edge| {
                edge.dbg();
            }
            std.debug.print(" }}", .{});
        }
    };

    const ParentPtr = struct {
        parent: *Node,
        idx: u16,

        fn into_node_or_leaf(self: ?ParentPtr) ?NodeOrLeaf {
            if (self) |ptr| {
                return .{ .node = ptr.parent };
            } else {
                return null;
            }
        }
    };

    const Leaf = struct {
        ally: std.mem.Allocator,
        level: usize = 0,
        parent: ?ParentPtr = null,

        len: u16 = 0,
        values: [CAPACITY]u32 = undefined,

        fn create(ally: std.mem.Allocator) !*Leaf {
            var leaf = try ally.create(Leaf);
            leaf.init(ally);

            return leaf;
        }

        fn dbg(self: *Leaf) void {
            const values = self.get_values();
            std.debug.print("{any}", .{values});
        }

        fn init(self: *Leaf, ally: std.mem.Allocator) void {
            self.* = Leaf{ .ally = ally };
        }

        fn destroy(self: *Leaf) void {
            self.ally.destroy(self);
        }

        fn push_value(self: *Leaf, value: u32) void {
            std.debug.assert(self.len < CAPACITY);
            var tmp = value;
            for (self.get_values()) |*val| {
                if (val.* < value) {
                    continue;
                }
                const t = val.*;
                val.* = tmp;
                tmp = t;
            }
            self.values[self.len] = tmp;
            self.len = self.len + 1;
        }

        const SplitResult = struct {
            // attached
            left: *Leaf,
            // lose value, previously attacked, must be inserted
            middle: u32,
            // free floating leaf, must be attached
            right: *Leaf,
        };

        fn split_at(self: *Leaf, value: u32) !SplitResult {
            var idx: u16 = 0;
            for (self.get_values(), 0..) |v, i| {
                idx = @intCast(u16, i);
                if (v > value) {
                    break;
                }
            }

            std.debug.assert(idx > 0 and idx < CAPACITY - 1);

            var new = try Leaf.create(self.ally);
            new.level = self.level;
            var middle: u32 = undefined;

            // take from right half
            if (idx > B) {
                new.len = self.len - (idx + 1);
                std.mem.copy(u32, &new.values, self.values[(idx + 1)..self.len]);

                middle = self.values[idx];

                self.len = idx;
                self.push_value(value);
            } else {
                // take from left half
                new.len = self.len - (idx);
                std.mem.copy(u32, &new.values, self.values[idx..self.len]);

                new.push_value(value);

                middle = self.values[idx - 1];

                self.len = idx - 1;
            }

            return .{ .left = self, .middle = middle, .right = new };
        }

        // returns null on success, or a split result which could not be merged
        // up because we are at the root node
        fn insert_value(self: *Leaf, value: u32) !?SplitResult {
            const leaf = self;

            if (leaf.len < CAPACITY) {
                std.debug.print("pushing value {} into leaf ", .{value});
                self.dbg();
                std.debug.print("\n", .{});

                leaf.push_value(value);
            } else {
                const split = try leaf.split_at(value);
                std.debug.print("splitting node ", .{});
                self.dbg();
                std.debug.print("into [", .{});
                split.left.dbg();
                std.debug.print(", {}, ", .{split.middle});
                split.right.dbg();
                std.debug.print("]\n", .{});

                if (leaf.parent) |parent| {
                    // parent can only throw split if it has no parent, so we just pass it back along to the top?
                    const result = try parent.parent.as_leaf().insert_value(split.middle);
                    parent.parent.insert_node(NodeOrLeaf.from_leaf(split.right));

                    if (result) |for_root| {
                        return for_root;
                    }
                } else {
                    return split;
                }
            }

            return null;
        }

        fn get_values(self: *Leaf) []u32 {
            const len = self.len;
            return self.values[0..len];
        }
    };
};

pub fn main() !void {
    // Prints to stderr (it's a shortcut based on `std.io.getStdErr()`)
    std.debug.print("All your {s} are belong to us.\n", .{"codebase"});

    // stdout is for the actual output of your application, for example if you
    // are implementing gzip, then only the compressed bytes should be sent to
    // stdout, not any debugging messages.
    const stdout_file = std.io.getStdOut().writer();
    var bw = std.io.bufferedWriter(stdout_file);
    const stdout = bw.writer();

    try stdout.print("Run `zig build test` to run the tests.\n", .{});

    try bw.flush(); // don't forget to flush!
}

test "btree leaf" {
    std.testing.refAllDeclsRecursive(BTree);
    std.testing.refAllDeclsRecursive(BTree.Leaf);

    var leaf = BTree.Leaf{ .ally = std.testing.allocator, .parent = null, .len = 2, .values = [_]u32{ 5, 6, undefined, undefined, undefined } };
    const values = leaf.get_values();

    std.debug.print("{?}\n", .{leaf});
    std.debug.print("{any}\n", .{values});
}

fn printValues(leaf: *BTree.Leaf) void {
    const values = leaf.get_values();
    std.debug.print("{any}\n", .{values});
}

test "leaf split" {
    std.debug.print("testing splitting\n", .{});

    var tree = BTree.create(std.testing.allocator);
    defer tree.destroy();
    try tree.insert(2);
    try tree.insert(4);
    try tree.insert(6);
    try tree.insert(3);
    try tree.insert(7);
    std.debug.print("before split:", .{});
    printValues(tree.root.?.as_leaf());

    const split = try tree.root.?.as_leaf().split_at(5);

    std.debug.print("after split:", .{});
    printValues(tree.root.?.as_leaf());

    std.debug.print("split: {?}\n", .{split});
    tree.ally.destroy(split.right);
    return error.SkipZigTest;
}

test "btree insert" {
    std.debug.print("testing insertion\n", .{});
    var tree = BTree.create(std.testing.allocator);
    defer tree.destroy();
    try tree.insert(10);
    try tree.insert(4);
    try tree.insert(6);
    try tree.insert(3);
    try tree.insert(9);
    try tree.insert(8);
    tree.dbg();
}

test "btree new" {
    std.debug.print("testing insertion\n", .{});
    var tree = BTree.create(std.testing.allocator);
    defer tree.destroy();
    try tree.insert(5);
    printValues(tree.root.?.as_leaf());
    try tree.insert(4);
    printValues(tree.root.?.as_leaf());
    try tree.insert(6);
    printValues(tree.root.?.as_leaf());
    try tree.insert(3);
    printValues(tree.root.?.as_leaf());
    try tree.insert(7);
    printValues(tree.root.?.as_leaf());
    //try tree.insert(8);
}

test "simple test" {
    var list = std.ArrayList(i32).init(std.testing.allocator);
    defer list.deinit(); // try commenting this out and see if zig detects the memory leak!
    try list.append(42);
    try std.testing.expectEqual(@as(i32, 42), list.pop());
}
